/*
 * drivers/sbus/audio/cs4231.c
 *
 * Copyright (C) 1996 Thomas K. Dyas (tdyas@noc.rutgers.edu)
 * Copyright (C) 1996 Derrick J Brashear (shadow@andrew.cmu.edu)
 * with substantial debugging help from Matthew Andrews (ma3d@andrew.cmu.edu)
 * and Chaskiel Grundman (cg2v+@andrew.cmu.edu)
 *
 * This is the lowlevel driver for the CS4231 audio chip found on some
 * sun4m and sun4u machines.
 */

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/errno.h>
#include <linux/interrupt.h>
#include <linux/malloc.h>
#include <asm/openprom.h>
#include <asm/oplib.h>
#include <asm/system.h>
#include <asm/irq.h>
#include <asm/io.h>
#include <asm/delay.h>
#include <asm/sbus.h>

#include "audio.h"
#include "cs4231.h"

/* Stolen for now from compat.h */
#ifndef MAX                             /* Usually found in <sys/param.h>. */
#define MAX(_a,_b)      ((_a)<(_b)?(_b):(_a))
#endif
#ifndef MIN                             /* Usually found in <sys/param.h>. */
#define MIN(_a,_b)      ((_a)<(_b)?(_a):(_b))
#endif

#define dprintk(x) printk x
#define MAX_DRIVERS 1
static struct sparcaudio_driver drivers[MAX_DRIVERS];
static int num_drivers;

static void cs4231_ready(struct sparcaudio_driver *drv) ;
static void cs4231_playintr(struct sparcaudio_driver *drv);
static int cs4231_recintr(struct sparcaudio_driver *drv);
static void cs4231_output_muted(struct sparcaudio_driver *drv, unsigned int value);
static void cs4231_mute(struct sparcaudio_driver *drv);
static void cs4231_pollinput(struct sparcaudio_driver *drv);
static int cs4231_attach(struct sparcaudio_driver *drv, 
			 struct linux_sbus_device *sdev);
static int cs4231_length_to_samplecount(struct audio_prinfo *thisdir, unsigned int length);
static void cs4231_getsamplecount(struct sparcaudio_driver *drv, unsigned int length, unsigned int value);

#define CHIP_READY udelay(100); cs4231_ready(drv); udelay(1000);

/* Enable cs4231 interrupts atomically. */
static __inline__ void cs4231_enable_interrupts(struct sparcaudio_driver *drv)
{
  struct cs4231_chip *cs4231_chip = (struct cs4231_chip *)drv->private;
  register unsigned long flags;

  if (cs4231_chip->status & CS_STATUS_INTS_ON)
    return;

  dprintk(("enabling interrupts\n"));
  save_and_cli(flags);
  cs4231_chip->regs->iar = 0xa;
  cs4231_chip->regs->idr = INTR_ON;
  restore_flags(flags);

  cs4231_chip->status |= CS_STATUS_INTS_ON;
}

/* Disable cs4231 interrupts atomically. */
static __inline__ void cs4231_disable_interrupts(struct sparcaudio_driver *drv)
{
  struct cs4231_chip *cs4231_chip = (struct cs4231_chip *)drv->private;
  register unsigned long flags;

  if (!(cs4231_chip->status & CS_STATUS_INTS_ON))
    return;

  dprintk(("disabling interrupts\n"));
  save_and_cli(flags);
  cs4231_chip->regs->iar = 0xa;
  cs4231_chip->regs->idr = INTR_OFF;
  restore_flags(flags);

  cs4231_chip->status &= ~CS_STATUS_INTS_ON;
}

static __inline__ void cs4231_enable_play(struct sparcaudio_driver *drv)
{
  struct cs4231_chip *cs4231_chip = (struct cs4231_chip *)drv->private;
  register unsigned long flags;

  dprintk(("enabling play\n"));
  save_and_cli(flags);
  cs4231_chip->regs->iar = 0x9;
  cs4231_chip->regs->idr |= PEN_ENABLE;
  restore_flags(flags);
}

static __inline__ void cs4231_disable_play(struct sparcaudio_driver *drv)
{
  struct cs4231_chip *cs4231_chip = (struct cs4231_chip *)drv->private;
  register unsigned long flags;

  dprintk(("disabling play\n"));
  save_and_cli(flags);
  cs4231_chip->regs->iar = 0x9;
  cs4231_chip->regs->idr &= PEN_DISABLE;
  restore_flags(flags);
}

static __inline__ void cs4231_enable_rec(struct sparcaudio_driver *drv)
{
  struct cs4231_chip *cs4231_chip = (struct cs4231_chip *)drv->private;
  register unsigned long flags;

  dprintk(("enabling rec\n"));
  save_and_cli(flags);
  cs4231_chip->regs->iar = 0x9;
  cs4231_chip->regs->idr |= CEN_ENABLE;
  restore_flags(flags);
}

static __inline__ void cs4231_disable_rec(struct sparcaudio_driver *drv)
{
  struct cs4231_chip *cs4231_chip = (struct cs4231_chip *)drv->private;
  register unsigned long flags;

  dprintk(("disabling rec\n"));
  save_and_cli(flags);
  cs4231_chip->regs->iar = 0x9;
  cs4231_chip->regs->idr &= CEN_DISABLE;
  restore_flags(flags);
}

static void cs4231_setup(struct sparcaudio_driver *drv)
{
  struct cs4231_chip *cs4231_chip = (struct cs4231_chip *)drv->private;

  cs4231_chip->regs->iar = IAR_AUTOCAL_BEGIN | 0x08;
  cs4231_chip->regs->idr = DEFAULT_DATA_FMAT;
  cs4231_chip->regs->iar = IAR_AUTOCAL_BEGIN | 0x1c;
  cs4231_chip->regs->idr = DEFAULT_DATA_FMAT;
  
  CHIP_READY
    
  cs4231_chip->status &= ~CS_STATUS_NEED_INIT;
}

/* Wait until the auto calibration process has finished */
static void
cs4231_ready(struct sparcaudio_driver *drv) 
{
  struct cs4231_chip *cs4231_chip = (struct cs4231_chip *)drv->private;
  unsigned int x = 0;

  cs4231_chip->regs->iar = IAR_AUTOCAL_END;
  while (cs4231_chip->regs->iar == IAR_NOT_READY && x <= CS_TIMEOUT) {
    x++;
  }

  x = 0;
  cs4231_chip->regs->iar = 0x0b;
  while (cs4231_chip->regs->idr == AUTOCAL_IN_PROGRESS && x <= CS_TIMEOUT) {
    x++;
  }
}

/* Audio interrupt handler. */
void cs4231_interrupt(int irq, void *dev_id, struct pt_regs *regs)
{
  struct sparcaudio_driver *drv = (struct sparcaudio_driver *)dev_id;
  struct cs4231_chip *cs4231_chip = (struct cs4231_chip *)drv->private;
  __u32 dummy;
  
  dprintk(("in cs4231_interrupt\n"));

  /* Clear the interrupt. */
  dummy = cs4231_chip->regs->dmacsr;
  cs4231_chip->regs->dmacsr = dummy;

#if 1
  dprintk(("csr is 0x%x\n", dummy));
#endif

  /* now go through and figure out what gets to claim the interrupt 
   * if anything since we may be doing shared interrupts 
   */

  if (dummy & CS_PLAY_INT) {
    if (dummy & CS_XINT_PNVA) {
      cs4231_chip->perchip_info.play.samples += 
	cs4231_length_to_samplecount(&(cs4231_chip->perchip_info.play), 
				     cs4231_chip->playlen);
      cs4231_playintr(drv);
    }
  }

  if (dummy & CS_CAPT_INT) {
    if (dummy & CS_XINT_CNVA) {
      cs4231_chip->perchip_info.record.samples += 
	cs4231_length_to_samplecount(&(cs4231_chip->perchip_info.record), 
				     cs4231_chip->reclen);
      cs4231_recintr(drv);
    }
  }
  if ((dummy & CS_XINT_CEMP) 
      && (cs4231_chip->perchip_info.record.active == 0)) 
    {
      cs4231_chip->perchip_info.record.active = 0;
    }
  if ((dummy & CS_XINT_EMPT) && (cs4231_chip->perchip_info.play.active == 0)) {
    cs4231_chip->regs->dmacsr |= (CS_PPAUSE);
    cs4231_disable_play(drv);
    cs4231_mute(drv);
    
    cs4231_getsamplecount(drv, cs4231_chip->playlen, 0);
    /* cleanup DMA */
    cs4231_chip->perchip_info.play.active = 0;
  }

  if (dummy & CS_GENL_INT) {
    /* If we get here we must be sharing an interrupt, but I haven't code 
       to handle this right now */
  }

  dprintk(("leaving cs4231_interrupt\n"));
}

/* Set output mute */
static void cs4231_output_muted(struct sparcaudio_driver *drv, unsigned int value)
{
  struct cs4231_chip *cs4231_chip = (struct cs4231_chip *)drv->private;
  dprintk(("in cs4231_output_muted: %d\n", value));
  if (!value) {
    cs4231_chip->regs->iar = 0x7;
    cs4231_chip->regs->idr &= OUTCR_UNMUTE;
    cs4231_chip->regs->iar = 0x6;
    cs4231_chip->regs->idr &= OUTCR_UNMUTE;
    cs4231_chip->perchip_info.output_muted = 0;
  } else {
    cs4231_chip->regs->iar = 0x7;
    cs4231_chip->regs->idr |= OUTCR_MUTE;
    cs4231_chip->regs->iar = 0x6;
    cs4231_chip->regs->idr |= OUTCR_MUTE;
    cs4231_chip->perchip_info.output_muted = 1;
  }
  dprintk(("leaving cs4231_output_muted\n"));
  return /*(cs4231_chip->perchip_info.output_muted)*/;
}

/* Set chip "output" port */
static unsigned int cs4231_out_port(struct sparcaudio_driver *drv, unsigned int value)
{
  struct cs4231_chip *cs4231_chip = (struct cs4231_chip *)drv->private;
  unsigned int retval = 0;

  dprintk(("in cs4231_out_port: %d\n", value));
  /* Aaaaaah! It's all coming so fast! Turn it all off, then selectively
   * enable things.
   */
  cs4231_chip->regs->iar = 0x1a;
  cs4231_chip->regs->idr |= MONO_IOCR_MUTE;
  cs4231_chip->regs->iar = 0x0a;
  cs4231_chip->regs->idr |= PINCR_LINE_MUTE;
  cs4231_chip->regs->idr |= PINCR_HDPH_MUTE;

  if (value & AUDIO_SPEAKER) {
    cs4231_chip->regs->iar = 0x1a;
    cs4231_chip->regs->idr &= ~MONO_IOCR_MUTE;
    retval |= AUDIO_SPEAKER;
  }

  if (value & AUDIO_HEADPHONE) {
   cs4231_chip->regs->iar = 0x0a;
   cs4231_chip->regs->idr &= ~PINCR_HDPH_MUTE;
   retval |= AUDIO_HEADPHONE;
  }

  if (value & AUDIO_LINE_OUT) {
    cs4231_chip->regs->iar = 0x0a;
    cs4231_chip->regs->idr &= ~PINCR_LINE_MUTE;
    retval |= AUDIO_LINE_OUT;
  }
  
  dprintk(("leaving cs4231_out_port\n"));
  return (retval);
}

/* Set chip "input" port */
static unsigned int cs4231_in_port(struct sparcaudio_driver *drv, unsigned int value)
{
  struct cs4231_chip *cs4231_chip = (struct cs4231_chip *)drv->private;
  unsigned int retval = 0;

  dprintk(("in cs4231_in_port: %d\n", value));

  /* You can have one and only one. This is probably wrong, but
   * appears to be how SunOS is doing it. Should be able to mix.
   */
  if (value & AUDIO_INTERNAL_CD_IN) {
    cs4231_chip->regs->iar = 0x1;
    cs4231_chip->regs->idr = CDROM_ENABLE(cs4231_chip->regs->idr);
    cs4231_chip->regs->iar = 0x0;
    cs4231_chip->regs->idr = CDROM_ENABLE(cs4231_chip->regs->idr);
    retval = AUDIO_INTERNAL_CD_IN;
  }
  if ((value & AUDIO_LINE_IN)) {
    cs4231_chip->regs->iar = 0x1;
    cs4231_chip->regs->idr = LINE_ENABLE(cs4231_chip->regs->idr);
    cs4231_chip->regs->iar = 0x0;
    cs4231_chip->regs->idr = LINE_ENABLE(cs4231_chip->regs->idr);
    retval = AUDIO_LINE_IN;
  } else if (value & AUDIO_MICROPHONE) {
    cs4231_chip->regs->iar = 0x1;
    cs4231_chip->regs->idr = MIC_ENABLE(cs4231_chip->regs->idr);
    cs4231_chip->regs->iar = 0x0;
    cs4231_chip->regs->idr = MIC_ENABLE(cs4231_chip->regs->idr);
    retval = AUDIO_MICROPHONE;
  } else if (value & AUDIO_ANALOG_LOOPBACK) {
    cs4231_chip->regs->iar = 0x1;
    cs4231_chip->regs->idr = OUTPUTLOOP_ENABLE(cs4231_chip->regs->idr);
    cs4231_chip->regs->iar = 0x0;
    cs4231_chip->regs->idr = OUTPUTLOOP_ENABLE(cs4231_chip->regs->idr);
    retval = AUDIO_ANALOG_LOOPBACK;
  }

  dprintk(("leaving cs4231_in_port\n"));
  return (retval);
}

/* Set chip "monitor" gain */
static unsigned int cs4231_monitor_gain(struct sparcaudio_driver *drv, unsigned int value)
{
  struct cs4231_chip *cs4231_chip = (struct cs4231_chip *)drv->private;
  int a = 0;

  dprintk(("in cs4231_monitor_gain: %d\n", value));

  /* This interpolation really sucks. The question is, be compatible 
   * with ScumOS/Sloaris or not?
   */
  a = CS4231_MON_MAX_ATEN - (value * (CS4231_MON_MAX_ATEN + 1) / (AUDIO_MAX_GAIN + 1));

  cs4231_chip->regs->iar = 0x0d;
  if (a >= CS4231_MON_MAX_ATEN) 
    cs4231_chip->regs->idr = LOOPB_OFF;
  else 
    cs4231_chip->regs->idr = ((a << 2) | LOOPB_ON);

  if (value == AUDIO_MAX_GAIN) return AUDIO_MAX_GAIN;

  dprintk(("leaving cs4231_monitor_gain\n"));

  return ((CS4231_MAX_DEV_ATEN - a) * (AUDIO_MAX_GAIN + 1) / (CS4231_MAX_DEV_ATEN + 1));
}

/* Set chip record gain */
static unsigned int cs4231_record_gain(struct sparcaudio_driver *drv, unsigned int value, unsigned char balance)
{
  struct cs4231_chip *cs4231_chip = (struct cs4231_chip *)drv->private;
  unsigned int tmp = 0, r, l, r_adj, l_adj;
  unsigned char old_gain;

  r = l = value;

  if (balance < AUDIO_MID_BALANCE) {
    r = MAX(0, (int)(value - ((AUDIO_MID_BALANCE - balance) << AUDIO_BALANCE_SHIFT)));
  } else if (balance > AUDIO_MID_BALANCE) {
    l = MAX(0, (int)(value - ((balance - AUDIO_MID_BALANCE) << AUDIO_BALANCE_SHIFT)));
  }

  l_adj = l * (CS4231_MAX_GAIN + 1) / (AUDIO_MAX_GAIN + 1);
  r_adj = r * (CS4231_MAX_GAIN + 1) / (AUDIO_MAX_GAIN + 1);
  
  cs4231_chip->regs->iar = 0x0;
  old_gain = cs4231_chip->regs->idr;
  cs4231_chip->regs->idr = RECGAIN_SET(old_gain, l_adj);
  cs4231_chip->regs->iar = 0x1;
  old_gain = cs4231_chip->regs->idr;
  cs4231_chip->regs->idr = RECGAIN_SET(old_gain, r_adj);
  
  if (l == value) {
    (l == 0) ? (tmp = 0) : (tmp = ((l_adj + 1) * AUDIO_MAX_GAIN) / (CS4231_MAX_GAIN + 1));
  } else if (r == value) {
    (r == 0) ? (tmp = 0) : (tmp = ((r_adj + 1) * AUDIO_MAX_GAIN) / (CS4231_MAX_GAIN + 1));
  }
  return (tmp);
}

/* Set chip play gain */
static unsigned int cs4231_play_gain(struct sparcaudio_driver *drv, unsigned int value, unsigned char balance)
{
  struct cs4231_chip *cs4231_chip = (struct cs4231_chip *)drv->private;
  unsigned int tmp = 0, r, l, r_adj, l_adj;
  unsigned char old_gain;

  dprintk(("in play_gain: %d %c\n", value, balance));
  r = l = value;
  if (balance < AUDIO_MID_BALANCE) {
    r = MAX(0, (int)(value - ((AUDIO_MID_BALANCE - balance) << AUDIO_BALANCE_SHIFT)));
  } else if (balance > AUDIO_MID_BALANCE) {
    l = MAX(0, (int)(value - ((balance - AUDIO_MID_BALANCE) << AUDIO_BALANCE_SHIFT)));
  }

  if (l == 0) {
    l_adj = CS4231_MAX_DEV_ATEN;
  } else {
    l_adj = CS4231_MAX_ATEN - (l * (CS4231_MAX_ATEN + 1) / (AUDIO_MAX_GAIN + 1));
  }
  if (r == 0) {
    r_adj = CS4231_MAX_DEV_ATEN;
  } else {
    r_adj = CS4231_MAX_ATEN - (r * (CS4231_MAX_ATEN + 1) / (AUDIO_MAX_GAIN + 1));
  }
  
  cs4231_chip->regs->iar = 0x6;
  old_gain = cs4231_chip->regs->idr;
  cs4231_chip->regs->idr = GAIN_SET(old_gain, l_adj);
  cs4231_chip->regs->iar = 0x7;
  old_gain = cs4231_chip->regs->idr;
  cs4231_chip->regs->idr = GAIN_SET(old_gain, r_adj);
  
  if ((value == 0) || (value == AUDIO_MAX_GAIN)) {
    tmp = value;
  } else {
    if (l == value) {
      tmp = ((CS4231_MAX_ATEN - l_adj) * (AUDIO_MAX_GAIN + 1) / (CS4231_MAX_ATEN + 1));
    } else if (r == value) {
      tmp = ((CS4231_MAX_ATEN - r_adj) * (AUDIO_MAX_GAIN + 1) / (CS4231_MAX_ATEN + 1));
    }
  }
  dprintk(("leaving cs4231_play_gain\n"));
  return (tmp);
}

/* Reset the audio chip to a sane state. */
static void cs4231_reset(struct sparcaudio_driver *drv)
{
  struct cs4231_chip *cs4231_chip = (struct cs4231_chip *)drv->private;

  dprintk(("in cs4231_reset\n"));

  cs4231_chip->regs->dmacsr = CS_CHIP_RESET;
  cs4231_chip->regs->dmacsr = 0x00;
  cs4231_chip->regs->dmacsr |= CS_CDC_RESET;
  
  udelay(20);
  
  cs4231_chip->regs->dmacsr &= ~(CS_CDC_RESET);
  cs4231_chip->regs->iar |= IAR_AUTOCAL_BEGIN;
  
  CHIP_READY
    
  cs4231_chip->regs->iar = IAR_AUTOCAL_BEGIN | 0x0c;
  cs4231_chip->regs->idr = MISC_IR_MODE2;
  cs4231_chip->regs->iar = IAR_AUTOCAL_BEGIN | 0x08;
  cs4231_chip->regs->idr = DEFAULT_DATA_FMAT; /* Ulaw */
  
  CHIP_READY
    
  cs4231_chip->regs->iar = IAR_AUTOCAL_BEGIN | 0x1c;
  cs4231_chip->regs->idr = DEFAULT_DATA_FMAT; /* Ulaw */
  
  CHIP_READY

  cs4231_chip->regs->iar = 0x19;
  /* see what we can turn on */
  if (cs4231_chip->regs->idr & CS4231A) {
    dprintk(("This is a CS4231A\n"));
    cs4231_chip->status |= CS_STATUS_REV_A;
  } else
    cs4231_chip->status &= ~CS_STATUS_REV_A;
  
  cs4231_chip->regs->iar = IAR_AUTOCAL_BEGIN | 0x10;
  cs4231_chip->regs->idr = OLB_ENABLE;
  
  cs4231_chip->regs->iar = IAR_AUTOCAL_BEGIN | 0x11;
  if (cs4231_chip->status & CS_STATUS_REV_A)
    cs4231_chip->regs->idr = (HPF_ON | XTALE_ON);
  else
    cs4231_chip->regs->idr = (HPF_ON);
  
  cs4231_chip->regs->iar = IAR_AUTOCAL_BEGIN | 0x1a;
  cs4231_chip->regs->idr = 0x00;
  
  /* Now set things up for defaults */
  cs4231_chip->perchip_info.play.gain = cs4231_play_gain(drv, CS4231_DEFAULT_PLAYGAIN, AUDIO_MID_BALANCE);
  cs4231_chip->perchip_info.record.gain = cs4231_record_gain(drv, CS4231_DEFAULT_RECGAIN, AUDIO_MID_BALANCE);

  cs4231_chip->perchip_info.record.port = cs4231_in_port(drv, AUDIO_MICROPHONE);
  cs4231_chip->perchip_info.play.port = cs4231_out_port(drv, AUDIO_SPEAKER);

  cs4231_chip->perchip_info.monitor_gain = cs4231_monitor_gain(drv, LOOPB_OFF);
  
  cs4231_chip->regs->iar = IAR_AUTOCAL_END;
  
  cs4231_ready(drv);
  
  cs4231_chip->regs->iar = IAR_AUTOCAL_BEGIN | 0x09;
  cs4231_chip->regs->idr &= ACAL_DISABLE;
  cs4231_chip->regs->iar = IAR_AUTOCAL_END;
  
  cs4231_ready(drv);

  cs4231_output_muted(drv, 0);

  dprintk(("leaving cs4231_reset\n"));
}

static void cs4231_mute(struct sparcaudio_driver *drv)
{
  struct cs4231_chip *cs4231_chip = (struct cs4231_chip *)drv->private;

  dprintk(("in cs4231_mute\n"));

  /* Supposedly I need to disable ACAL here, then turn it back on after */
  /* Makes no sense */
  if (!(cs4231_chip->status & CS_STATUS_REV_A)) {
    cs4231_chip->regs->iar = IAR_AUTOCAL_BEGIN;
    udelay(100);
    cs4231_chip->regs->iar = IAR_AUTOCAL_END;
    CHIP_READY
  }

  dprintk(("leaving cs4231_mute\n"));
}

static int cs4231_length_to_samplecount(struct audio_prinfo *thisdir, unsigned int length)
{
  unsigned int count;

  if (thisdir->channels == 2)
    count = (length/2);
  else 
    count = length;
  
  if (thisdir->encoding == AUDIO_ENCODING_LINEAR)
    count = (count/2);
  
  return count;
}

static void cs4231_getsamplecount(struct sparcaudio_driver *drv, unsigned int length, unsigned int value)
{
  struct cs4231_chip *cs4231_chip = (struct cs4231_chip *)drv->private;
  struct audio_prinfo *thisdir;
  unsigned int count, nextcount, curcount;

  if (value == 1) /* record */ 
    {
      thisdir = &cs4231_chip->perchip_info.record;
      curcount = 
	cs4231_length_to_samplecount(thisdir, cs4231_chip->regs->dmacc);
      nextcount = 
	cs4231_length_to_samplecount(thisdir, cs4231_chip->regs->dmacnc);
    }
  else /* play */
    {
      thisdir = &cs4231_chip->perchip_info.play;
      curcount = 
	cs4231_length_to_samplecount(thisdir, cs4231_chip->regs->dmapc);
      nextcount = 
	cs4231_length_to_samplecount(thisdir, cs4231_chip->regs->dmapnc);
    }
  count = thisdir->samples;
  length = cs4231_length_to_samplecount(thisdir, length);
  thisdir->samples = ((count - nextcount) + (length - curcount));
}

static int cs4231_open(struct inode * inode, struct file * file, struct sparcaudio_driver *drv)
{	
  struct cs4231_chip *cs4231_chip = (struct cs4231_chip *)drv->private;

  dprintk(("in cs4231_open\n"));

  /* Set the default audio parameters. */
  
  cs4231_chip->perchip_info.play.sample_rate = CS4231_RATE;
  cs4231_chip->perchip_info.play.channels = CS4231_CHANNELS;
  cs4231_chip->perchip_info.play.precision = CS4231_PRECISION;
  cs4231_chip->perchip_info.play.encoding = AUDIO_ENCODING_ULAW;
  
  cs4231_chip->perchip_info.record.sample_rate = CS4231_RATE;
  cs4231_chip->perchip_info.record.channels = CS4231_CHANNELS;
  cs4231_chip->perchip_info.record.precision = CS4231_PRECISION;
  cs4231_chip->perchip_info.record.encoding = AUDIO_ENCODING_ULAW;
  
  cs4231_ready(drv);
  
  cs4231_chip->status |= CS_STATUS_NEED_INIT;
  
  CHIP_READY
    
  MOD_INC_USE_COUNT;
  
  dprintk(("leaving cs4231_open\n"));
  return 0;
}

static void cs4231_release(struct inode * inode, struct file * file, struct sparcaudio_driver *drv)
{
  struct cs4231_chip *cs4231_chip = (struct cs4231_chip *)drv->private;

  dprintk(("in cs4231_release\n"));
  /* zero out any info about what data we have as well */

  if (!cs4231_chip->perchip_info.play.open && !cs4231_chip->perchip_info.record.open && (cs4231_chip->status & CS_STATUS_INIT_ON_CLOSE)) {
    cs4231_reset(drv);
    cs4231_chip->status &= ~CS_STATUS_INIT_ON_CLOSE;
  }

  MOD_DEC_USE_COUNT;
  dprintk(("leaving cs4231_release\n"));
}

static void cs4231_playintr(struct sparcaudio_driver *drv)
{
  struct cs4231_chip *cs4231_chip = (struct cs4231_chip *)drv->private;
  unsigned int length = 0;

  dprintk(("in playintr: ptr %ld len %ld\n", cs4231_chip->output_ptr, cs4231_chip->output_count));
  if (cs4231_chip->playlen == 0)
    cs4231_chip->playlen = cs4231_chip->output_count;
  if (cs4231_chip->output_ptr && cs4231_chip->output_count > 0) {
    /* Calculate number of samples here */
    cs4231_chip->playlen = length;
#if 0
    if (cs4231_chip->output_count > 1024) 
      cs4231_chip->regs->dmapnc = 1024;
    else
#endif
      cs4231_chip->regs->dmapnc = cs4231_chip->output_count;

    cs4231_chip->regs->dmapnva = (__u32) cs4231_chip->output_ptr;

#if 0
    if (cs4231_chip->output_count > 1024) {
      cs4231_chip->output_ptr+=1024;
      cs4231_chip->output_count-=1024;
    } else {
      cs4231_chip->output_ptr+=cs4231_chip->output_count;
      cs4231_chip->output_count=0;
    }

    if (cs4231_chip->output_count == 0) {
#endif
      cs4231_chip->output_count = 0;
      cs4231_chip->output_ptr = NULL;
      sparcaudio_output_done(drv);
#if 0
    }
#endif
  }
  return; 
}

static void cs4231_recmute(int fmt)
{
  switch (fmt) {
  case AUDIO_ENCODING_LINEAR:
    /* Insert 0x00 from "here" to end of data stream */
    break;
  case AUDIO_ENCODING_ALAW:
    /* Insert 0xd5 from "here" to end of data stream */
    break;
  case AUDIO_ENCODING_ULAW:
    /* Insert 0xff from "here" to end of data stream */
    break;
  }
}

static int cs4231_recintr(struct sparcaudio_driver *drv)
{
  struct cs4231_chip *cs4231_chip = (struct cs4231_chip *)drv->private;

  cs4231_recmute(cs4231_chip->perchip_info.record.encoding);

  if (cs4231_chip->perchip_info.record.active == 0) {
    cs4231_pollinput(drv);
    cs4231_disable_rec(drv);    
  }
  /* Read the next byte of incoming data. */
#if 0
  if (cs4231_chip->input_ptr && cs4231_chip->input_count > 0) {
    cs4231_chip->regs->dmacnva = dma_handle;
    cs4231_chip->regs->dmacnc = length;
    cs4231_chip->input_ptr++;
    cs4231_chip->input_count--;
    
    /* Done with the buffer? Notify the midlevel driver. */
    if (cs4231_chip->input_count == 0) {
      cs4231_chip->input_ptr = NULL;
      cs4231_chip->input_count = 0;
      sparcaudio_input_done(drv);
    }
  }
#endif
  return 1;
}

static void cs4231_start_output(struct sparcaudio_driver *drv, __u8 * buffer, unsigned long count)
{
  struct cs4231_chip *cs4231_chip = (struct cs4231_chip *)drv->private;

  cs4231_chip->output_ptr = buffer;
  cs4231_chip->output_count = count;

  dprintk(("in cs4231_start_output\n"));
  if (cs4231_chip->perchip_info.play.active || (cs4231_chip->perchip_info.play.pause))
    return;

  cs4231_ready(drv);

  if (cs4231_chip->status & CS_STATUS_NEED_INIT)
    cs4231_setup(drv);

  if (!cs4231_chip->perchip_info.play.pause) 
    {
      dprintk(("clearing pause\n"));
      cs4231_chip->regs->dmacsr &= ~CS_XINT_PLAY;
      cs4231_chip->regs->dmacsr &= ~CS_PPAUSE;
      cs4231_chip->regs->dmacsr |= CS_PLAY_SETUP;
      cs4231_playintr(drv);
      /*      cs4231_chip->regs->dmacsr |= CS_PLAY_SETUP;*/
      cs4231_ready(drv);
      cs4231_enable_play(drv);
    }
  cs4231_chip->perchip_info.play.active = 1;
  dprintk(("leaving cs4231_start_output\n"));
}

static void cs4231_stop_output(struct sparcaudio_driver *drv)
{
  struct cs4231_chip *cs4231_chip = (struct cs4231_chip *)drv->private;

  dprintk(("in cs4231_stop_output\n"));
  cs4231_chip->output_ptr = NULL;
  cs4231_chip->output_count = 0;
  cs4231_chip->perchip_info.play.active = 0;
  cs4231_chip->regs->dmacsr |= (CS_PPAUSE);
  dprintk(("leaving cs4231_stop_output\n"));
}

static void cs4231_pollinput(struct sparcaudio_driver *drv)
{
  struct cs4231_chip *cs4231_chip = (struct cs4231_chip *)drv->private;
  int x = 0;

  while (!(cs4231_chip->regs->dmacsr & CS_XINT_COVF) && x <= CS_TIMEOUT) {
    x++;
  }
  cs4231_chip->regs->dmacsr |= CS_XINT_CEMP;
}

static void cs4231_start_input(struct sparcaudio_driver *drv, __u8 * buffer, unsigned long count)
{
  struct cs4231_chip *cs4231_chip = (struct cs4231_chip *)drv->private;

  if (cs4231_chip->perchip_info.record.active || (cs4231_chip->perchip_info.record.pause))
    return;

  cs4231_ready(drv);

  if (cs4231_chip->status & CS_STATUS_NEED_INIT)
    cs4231_setup(drv);

  if (!cs4231_chip->perchip_info.record.pause)
    {
      /* init dma foo here */
      cs4231_chip->regs->dmacsr &= ~CS_XINT_CAPT;
      cs4231_chip->regs->dmacsr &= ~CS_CPAUSE;
      cs4231_recintr(drv);
      cs4231_chip->regs->dmacsr |= CS_CAPT_SETUP;
      cs4231_enable_rec(drv);
    }
  cs4231_chip->perchip_info.record.active = 1;
}

static void cs4231_stop_input(struct sparcaudio_driver *drv)
{
  struct cs4231_chip *cs4231_chip = (struct cs4231_chip *)drv->private;

  cs4231_chip->perchip_info.record.active = 0;
  cs4231_chip->regs->dmacsr |= (CS_CPAUSE);

  cs4231_pollinput(drv);

  /* need adjust the end pointer, process the input, and clean up the dma */

  cs4231_disable_rec(drv);
}

static void cs4231_audio_getdev(struct sparcaudio_driver *drv,
                                 audio_device_t * audinfo)
{
        strncpy(audinfo->name, "cs4231", sizeof(audinfo->name) - 1);
        strncpy(audinfo->version, "x", sizeof(audinfo->version) - 1);
        strncpy(audinfo->config, "audio", sizeof(audinfo->config) - 1);
}


static void cs4231_loopback(struct sparcaudio_driver *drv, unsigned int value)
{
  struct cs4231_chip *cs4231_chip = (struct cs4231_chip *)drv->private;

  cs4231_chip->regs->iar = 0xd;
  cs4231_chip->regs->idr = (value ? LOOPB_ON : 0);
}

static int cs4231_ioctl(struct inode * inode, struct file * file,
			unsigned int cmd, unsigned long arg, 
			struct sparcaudio_driver *drv)
{
  struct cs4231_chip *cs4231_chip = (struct cs4231_chip *)drv->private;
  int retval = 0;
  
  switch (cmd) {
  case AUDIO_DIAG_LOOPBACK:
    cs4231_chip->status |= CS_STATUS_INIT_ON_CLOSE;
    cs4231_loopback(drv, (unsigned int)arg);
    break;
  default:
    retval = -EINVAL;
  }

  return retval;
}

/* There will also be a handler for getinfo and setinfo */

static struct sparcaudio_operations cs4231_ops = {
	cs4231_open,
	cs4231_release,
	cs4231_ioctl,
	cs4231_start_output,
	cs4231_stop_output,
	cs4231_start_input,
        cs4231_stop_input,
	cs4231_audio_getdev,
};

/* Probe for the cs4231 chip and then attach the driver. */
#ifdef MODULE
int init_module(void)
#else
__initfunc(int cs4231_init(void))
#endif
{
  struct linux_sbus *bus;
  struct linux_sbus_device *sdev;
  
  num_drivers = 0;
  
  /* Probe each SBUS for cs4231 chips. */
  for_all_sbusdev(sdev,bus) {
    if (!strcmp(sdev->prom_name, "SUNW,CS4231")) {
      /* Don't go over the max number of drivers. */
      if (num_drivers >= MAX_DRIVERS)
	continue;
      
      if (cs4231_attach(&drivers[num_drivers], sdev) == 0)
	num_drivers++;
    }
  }
  
  /* Only return success if we found some cs4231 chips. */
  return (num_drivers > 0) ? 0 : -EIO;
}

/* Attach to an cs4231 chip given its PROM node. */
static int cs4231_attach(struct sparcaudio_driver *drv, 
			 struct linux_sbus_device *sdev)
{
  struct linux_prom_irqs irq;
  struct cs4231_chip *cs4231_chip;
  int err;

  /* Allocate our private information structure. */
  drv->private = kmalloc(sizeof(struct cs4231_chip), GFP_KERNEL);
  if (!drv->private)
    return -ENOMEM;

  /* Point at the information structure and initialize it. */
  drv->ops = &cs4231_ops;
  cs4231_chip = (struct cs4231_chip *)drv->private;
  cs4231_chip->input_ptr = cs4231_chip->output_ptr = NULL;
  cs4231_chip->input_count = cs4231_chip->output_count = 0;

  /* Map the registers into memory. */
  prom_apply_sbus_ranges(sdev->my_bus, &sdev->reg_addrs[0], 
			 sdev->num_registers, sdev);

  cs4231_chip->regs_size = sdev->reg_addrs[0].reg_size;

  cs4231_chip->regs = sparc_alloc_io(sdev->reg_addrs[0].phys_addr, 0,
				     sdev->reg_addrs[0].reg_size, 
				     "cs4231", sdev->reg_addrs[0].which_io, 
				     0);

  if (!cs4231_chip->regs) {
    printk(KERN_ERR "cs4231: could not allocate registers\n");
    kfree(drv->private);
    return -EIO;
  }

  /* Attach the interrupt handler to the audio interrupt. */
  prom_getproperty(sdev->prom_node, "intr", (char *)&irq, sizeof(irq));

  if (irq.pri < 0) {
    sparc_free_io(cs4231_chip->regs, cs4231_chip->regs_size);
    kfree(drv->private);
    return -EIO;
  }

  cs4231_chip->irq = irq.pri;

  /* Not sure if these are ever a shared interrupt so we just deal */
  if (request_irq(cs4231_chip->irq, cs4231_interrupt, SA_INTERRUPT|SA_SHIRQ, "cs4231", drv) == -EBUSY) 
    request_irq(cs4231_chip->irq, cs4231_interrupt, SA_INTERRUPT, "cs4231", drv);
  enable_irq(cs4231_chip->irq);

  cs4231_enable_interrupts(drv);

  /* Reset the audio chip. */
  cs4231_reset(drv);

  /* Register ourselves with the midlevel audio driver. */
  err = register_sparcaudio_driver(drv);

  if (err < 0) {
    printk(KERN_ERR "cs4231: unable to register\n");
    disable_irq(cs4231_chip->irq);
    free_irq(cs4231_chip->irq, drv);
    sparc_free_io(cs4231_chip->regs, cs4231_chip->regs_size);
    kfree(drv->private);
    return -EIO;
  }

  cs4231_chip->perchip_info.play.active = 
    cs4231_chip->perchip_info.play.pause = 0;

  cs4231_chip->perchip_info.record.active = 
    cs4231_chip->perchip_info.record.pause = 0;

  /* Announce the hardware to the user. */
  printk(KERN_INFO "cs4231%c at 0x%lx irq %d\n",
	 (cs4231_chip->status & CS_STATUS_REV_A) ? 'a' : ' ', 
	 (unsigned long)cs4231_chip->regs, cs4231_chip->irq);
  
  /* Success! */
  return 0;
}

#ifdef MODULE
/* Detach from an cs4231 chip given the device structure. */
static void cs4231_detach(struct sparcaudio_driver *drv)
{
        struct cs4231_chip *info = (struct cs4231_chip *)drv->private;

	cs4231_disable_interrupts(drv);
        unregister_sparcaudio_driver(drv);
        disable_irq(info->irq);
        free_irq(info->irq, drv);
        sparc_free_io(info->regs, info->regs_size);
        kfree(drv->private);
}

void cleanup_module(void)
{
        register int i;

        for (i = 0; i < num_drivers; i++) {
                cs4231_detach(&drivers[i]);
                num_drivers--;
        }
}
#endif
